/* NSGA-II routine (implementation of the 'main' function) */

# include <stdio.h>
# include <stdlib.h>
# include <math.h>
# include <unistd.h>

# include "global.h"
# include "rand.h"

int nreal;
int nbin;
int nobj;
int ncon;
int popsize;
double pcross_real;
double pcross_bin;
double pmut_real;
double pmut_bin;
double eta_c;
double eta_m;
int ngen;
int nbinmut;
int nrealmut;
int nbincross;
int nrealcross;
int *nbits;
double *min_realvar;
double *max_realvar;
double *min_binvar;
double *max_binvar;
int bitlength;
int choice;
int obj1;
int obj2;
int obj3;
int angle1;
int angle2;
int efficiency; /* 0: Pareto efficiency
			       1: equitable efficiency
			       2: variable domination: minimum
			       3: variable domination: nondmination*/
int restr; /* 0: standard NSGA II,
			  1: Geoffrion bound,
			  2: tradeoff bound
			  3: *
			  4: guided domination approach
			  5: variable domination structure: minimum
			  6: variable domination structure: nondomination */
int app_restr_all_fronts; /* create subfront from restriction for every front in every generation (app_restr_all_fronts==1)
							 or just for first front if front size exceeds popsize (app_restr_all_fronts==0) */
int troff_i; /* objective function i of tradeoff T_ij */
			 /* if (restr != 2) input has to be provided but has no influence*/
int troff_j; /* objective function j of tradeoff T_ij */
			 /* if (restr != 2) input has to be provided but has no influence*/
double delta; /* * */

double m_bound; /* Geoffrion/tradeoff bound */
				/* if (restr != 1) input has to be provided but has no influence*/
double gamma; /* cone paramter, see definition of Bishop-Phelps cones */
			  /* if (restr != (5||6)) input has to be provided but has no influence*/
double *ideal; /* components of the ideal vector; number of components == nobj */
			   /* if (restr != (5||6)) input has to be provided but has no influence*/
int runs; /* number of program runs */


int main (int argc, char **argv)
{
    int i;
    int runnr; /* current run number */
    char fname[32]; /* filenames of output files */
    FILE *fpt1;
    FILE *fpt2;
    FILE *fpt3;
    FILE *fpt4;
    FILE *fpt5;
    FILE *gp;
    population *parent_pop;
    population *child_pop;
    population *mixed_pop;
    population *f0_pop;
    if (argc<2)
    {
        printf("\n Usage ./nsga2r random_seed \n");
        exit(1);
    }
    seed = (double)atof(argv[1]);
    if (seed<=0.0 || seed>=1.0)
    {
        printf("\n Entered seed value is wrong, seed value must be in (0,1) \n");
        exit(1);
    }


    /* ----------------------- Read parameters ----------------------------------------------*/
    printf("\n Enter the problem relevant and algorithm relevant parameters ... ");
    printf("\n Enter the population size (a multiple of 4) : ");
    scanf("%d",&popsize);
    if (popsize<4 || (popsize%4)!= 0)
    {
        printf("\n population size read is : %d",popsize);
        printf("\n Wrong population size entered, hence exiting \n");
        exit (1);
    }
    printf("\n Enter the number of generations : ");
    scanf("%d",&ngen);
    if (ngen<1)
    {
        printf("\n number of generations read is : %d",ngen);
        printf("\n Wrong nuber of generations entered, hence exiting \n");
        exit (1);
    }
    printf("\n Enter the number of objectives : ");
    scanf("%d",&nobj);
    if (nobj<1)
    {
        printf("\n number of objectives entered is : %d",nobj);
        printf("\n Wrong number of objectives entered, hence exiting \n");
        exit (1);
    }
    printf("\n Enter the number of constraints : ");
    scanf("%d",&ncon);
    if (ncon<0)
    {
        printf("\n number of constraints entered is : %d",ncon);
        printf("\n Wrong number of constraints enetered, hence exiting \n");
        exit (1);
    }
    printf("\n Enter the number of real variables : ");
    scanf("%d",&nreal);
    if (nreal<0)
    {
        printf("\n number of real variables entered is : %d",nreal);
        printf("\n Wrong number of variables entered, hence exiting \n");
        exit (1);
    }
    if (nreal != 0)
    {
        min_realvar = (double *)malloc(nreal*sizeof(double));
        max_realvar = (double *)malloc(nreal*sizeof(double));
        for (i=0; i<nreal; i++)
        {
            printf ("\n Enter the lower limit of real variable %d : ",i+1);
            scanf ("%lf",&min_realvar[i]);
            printf ("\n Enter the upper limit of real variable %d : ",i+1);
            scanf ("%lf",&max_realvar[i]);
            if (max_realvar[i] <= min_realvar[i])
            {
                printf("\n Wrong limits entered for the min and max bounds of real variable, hence exiting \n");
                exit(1);
            }
        }
        printf ("\n Enter the probability of crossover of real variable (0.6-1.0) : ");
        scanf ("%lf",&pcross_real);
        if (pcross_real<0.0 || pcross_real>1.0)
        {
            printf("\n Probability of crossover entered is : %e",pcross_real);
            printf("\n Entered value of probability of crossover of real variables is out of bounds, hence exiting \n");
            exit (1);
        }
        printf ("\n Enter the probablity of mutation of real variables (1/nreal) : ");
        scanf ("%lf",&pmut_real);
        if (pmut_real<0.0 || pmut_real>1.0)
        {
            printf("\n Probability of mutation entered is : %e",pmut_real);
            printf("\n Entered value of probability of mutation of real variables is out of bounds, hence exiting \n");
            exit (1);
        }
        printf ("\n Enter the value of distribution index for crossover (5-20): ");
        scanf ("%lf",&eta_c);
        if (eta_c<=0)
        {
            printf("\n The value entered is : %e",eta_c);
            printf("\n Wrong value of distribution index for crossover entered, hence exiting \n");
            exit (1);
        }
        printf ("\n Enter the value of distribution index for mutation (5-50): ");
        scanf ("%lf",&eta_m);
        if (eta_m<=0)
        {
            printf("\n The value entered is : %e",eta_m);
            printf("\n Wrong value of distribution index for mutation entered, hence exiting \n");
            exit (1);
        }
    }
    printf("\n Enter the number of binary variables : ");
    scanf("%d",&nbin);
    if (nbin<0)
    {
        printf ("\n number of binary variables entered is : %d",nbin);
        printf ("\n Wrong number of binary variables entered, hence exiting \n");
        exit(1);
    }
    if (nbin != 0)
    {
        nbits = (int *)malloc(nbin*sizeof(int));
        min_binvar = (double *)malloc(nbin*sizeof(double));
        max_binvar = (double *)malloc(nbin*sizeof(double));
        for (i=0; i<nbin; i++)
        {
            printf ("\n Enter the number of bits for binary variable %d : ",i+1);
            scanf ("%d",&nbits[i]);
            if (nbits[i] < 1)
            {
                printf("\n Wrong number of bits for binary variable entered, hence exiting");
                exit(1);
            }
            printf ("\n Enter the lower limit of binary variable %d : ",i+1);
            scanf ("%lf",&min_binvar[i]);
            printf ("\n Enter the upper limit of binary variable %d : ",i+1);
            scanf ("%lf",&max_binvar[i]);
            if (max_binvar[i] <= min_binvar[i])
            {
                printf("\n Wrong limits entered for the min and max bounds of binary variable entered, hence exiting \n");
                exit(1);
            }
        }
        printf ("\n Enter the probability of crossover of binary variable (0.6-1.0): ");
        scanf ("%lf",&pcross_bin);
        if (pcross_bin<0.0 || pcross_bin>1.0)
        {
            printf("\n Probability of crossover entered is : %e",pcross_bin);
            printf("\n Entered value of probability of crossover of binary variables is out of bounds, hence exiting \n");
            exit (1);
        }
        printf ("\n Enter the probability of mutation of binary variables (1/nbits): ");
        scanf ("%lf",&pmut_bin);
        if (pmut_bin<0.0 || pmut_bin>1.0)
        {
            printf("\n Probability of mutation entered is : %e",pmut_bin);
            printf("\n Entered value of probability  of mutation of binary variables is out of bounds, hence exiting \n");
            exit (1);
        }
    }
    if (nreal==0 && nbin==0)
    {
        printf("\n Number of real as well as binary variables, both are zero, hence exiting \n");
        exit(1);
    }
    choice=0;
    printf("\n Do you want to use gnuplot to display the results realtime (0 for NO) (1 for yes) : ");
    scanf("%d",&choice);
    if (choice!=0 && choice!=1)
    {
        printf("\n Entered the wrong choice, hence exiting, choice entered was %d\n",choice);
        exit(1);
    }
    if (choice==1)
    {
        gp = popen(GNUPLOT_COMMAND,"w");
        if (gp==NULL)
        {
            printf("\n Could not open a pipe to gnuplot, check the definition of GNUPLOT_COMMAND in file global.h\n");
            printf("\n Edit the string to suit your system configuration and rerun the program\n");
            exit(1);
        }
        if (nobj==2)
        {
            printf("\n Enter the objective for X axis display : ");
            scanf("%d",&obj1);
            if (obj1<1 || obj1>nobj)
            {
                printf("\n Wrong value of X objective entered, value entered was %d\n",obj1);
                exit(1);
            }
            printf("\n Enter the objective for Y axis display : ");
            scanf("%d",&obj2);
            if (obj2<1 || obj2>nobj)
            {
                printf("\n Wrong value of Y objective entered, value entered was %d\n",obj2);
                exit(1);
            }
            obj3 = -1;
        }
        else
        {
            printf("\n #obj > 2, 2D display or a 3D display ?, enter 2 for 2D and 3 for 3D :");
            scanf("%d",&choice);
            if (choice!=2 && choice!=3)
            {
                printf("\n Entered the wrong choice, hence exiting, choice entered was %d\n",choice);
                exit(1);
            }
            if (choice==2)
            {
                printf("\n Enter the objective for X axis display : ");
                scanf("%d",&obj1);
                if (obj1<1 || obj1>nobj)
                {
                    printf("\n Wrong value of X objective entered, value entered was %d\n",obj1);
                    exit(1);
                }
                printf("\n Enter the objective for Y axis display : ");
                scanf("%d",&obj2);
                if (obj2<1 || obj2>nobj)
                {
                    printf("\n Wrong value of Y objective entered, value entered was %d\n",obj2);
                    exit(1);
                }
                obj3 = -1;
            }
            else
            {
                printf("\n Enter the objective for X axis display : ");
                scanf("%d",&obj1);
                if (obj1<1 || obj1>nobj)
                {
                    printf("\n Wrong value of X objective entered, value entered was %d\n",obj1);
                    exit(1);
                }
                printf("\n Enter the objective for Y axis display : ");
                scanf("%d",&obj2);
                if (obj2<1 || obj2>nobj)
                {
                    printf("\n Wrong value of Y objective entered, value entered was %d\n",obj2);
                    exit(1);
                }
                printf("\n Enter the objective for Z axis display : ");
                scanf("%d",&obj3);
                if (obj3<1 || obj3>nobj)
                {
                    printf("\n Wrong value of Z objective entered, value entered was %d\n",obj3);
                    exit(1);
                }
                printf("\n You have chosen 3D display, hence location of eye required \n");
                printf("\n Enter the first angle (an integer in the range 0-180) (if not known, enter 60) :");
                scanf("%d",&angle1);
                if (angle1<0 || angle1>180)
                {
                    printf("\n Wrong value for first angle entered, hence exiting \n");
                    exit(1);
                }
                printf("\n Enter the second angle (an integer in the range 0-360) (if not known, enter 30) :");
                scanf("%d",&angle2);
                if (angle2<0 || angle2>360)
                {
                    printf("\n Wrong value for second angle entered, hence exiting \n");
                    exit(1);
                }
            }
        }
    }

    printf("\n Enter kind of efficiency (0: Pareto, 1: equitable) :");
	scanf("%d",&efficiency);
	if (efficiency<0 || efficiency>3)
	{
		printf("\n Efficiency read is : %d", efficiency);
		printf("\n Wrong efficiency entered, hence exiting \n");
		exit (1);
	}
	printf("\n Efficiency entered : %d", efficiency);

    printf("\n Enter kind of restriction :");
	scanf("%d",&restr);
	if (restr<0 || restr>6)
	{
		printf("\n Restriction read is : %d", restr);
		printf("\n Wrong restriction entered, hence exiting \n");
		exit (1);
	}
	printf("\n Restriction entered : %d", restr);

	printf("\n Shall restriction be applied to every front (enter 0 or 1) :");
	/* create subfront from restriction for every front in every generation (app_restr_all_fronts==1)
	   or just for first front if front size exceeds popsize (app_restr_all_fronts==0) */
	scanf("%d",&app_restr_all_fronts);
	if (app_restr_all_fronts<0 || app_restr_all_fronts>1)
	{
		printf("\n app_restr_all_fronts read is : %d", restr);
		printf("\n Wrong app_restr_all_fronts entered, hence exiting \n");
		exit (1);
	}
	printf("\n app_restr_all_fronts entered : %d", app_restr_all_fronts);

	printf("\n Enter objective function i for tradeoff T_ij (0<=i<nobj) :"); /* if (restr != 2) input has to be provided but has no influence*/
	scanf("%d",&troff_i);
	if (restr==2 && (troff_i<0 || troff_i>nobj))
	{
		printf("\n Objective function read is : %d", troff_i);
		printf("\n Wrong restriction entered, hence exiting \n");
		exit (1);
	}
	printf("\n Objective function entered : %d", troff_i);

	printf("\n Enter objective function j for tradeoff T_ij :"); /* if (restr != 2) input has to be provided but has no influence*/
	scanf("%d",&troff_j);
	if (restr==2 && (troff_j<0 || troff_j>nobj || troff_i==troff_j))
	{
		printf("\n Objective function read is : %d", troff_j);
		printf("\n Wrong restriction entered, hence exiting \n");
		exit (1);
	}
	printf("\n Objective function entered : %d", troff_j);

	printf("\n Enter delta (>=0) :"); /* if (restr != 3) input has to be provided but has no influence*/
	scanf("%lf",&delta);
	if (delta<0)
	{
		printf("\n delta read is : %f", delta);
		printf("\n Wrong delta entered, hence exiting \n");
		exit (1);
	}
	printf("\n delta entered : %f", delta);

    printf("\n Enter Geoffrion/tradeoff bound (>0) :");
    scanf("%lf",&m_bound);
	if (m_bound<=0)
	{
		printf("\n Geoffrion/tradeoff bound M read is : %f", m_bound);
		printf("\n Wrong Geoffrion/tradeoff bound entered, hence exiting \n");
		exit (1);
	}
    printf("\n Geoffrion Bound entered : %f", m_bound);

    printf("\n Enter cone parameter gamma (0 <= gamma <= 1) :");
	scanf("%lf",&gamma);
	if (gamma<0 || gamma>1)
	{
		printf("\n gamma read is : %f", gamma);
		printf("\n Wrong gamma entered, hence exiting \n");
		exit (1);
	}
	printf("\n gamma entered : %f", gamma);

	/* Allocate memory for ideal vector and read components */
	ideal = (double *) malloc(nobj * sizeof(double));
	if (ideal == NULL)
	{
		printf("\n Memory Allocation for ideal failed, hence exiting \n");
		exit(1);
	}
	for (i=0; i<nobj; ++i)
	{
		printf("\n Enter component No %d of ideal vector :", i);
		scanf("%lf", ideal+i);
	}

    printf("\n Enter number of program runs (>=1) :");
	scanf("%d",&runs);
	if (runs<1)
	{
		printf("\n Number of program runs read is : %d", runs);
		printf("\n Wrong number of program runs entered, hence exiting \n");
		exit (1);
	}
	printf("\n Number of program runs entered : %d", runs);

    printf("\n Input data successfully entered, now performing initialization \n");


    /* ---------------------------------------------- Main loop ------------------------------------------- */
    randomize();
	for (runnr = 1; runnr<=runs; ++runnr)
	{
		if (runnr !=1 && choice==1)
		{
			gp = popen(GNUPLOT_COMMAND,"w");
		}
		sprintf(fname, "output/initial_pop_run%d.out", runnr);
		fpt1 = fopen(fname, "w");
		sprintf(fname, "output/final_pop_run%d.out", runnr);
		fpt2 = fopen(fname, "w");
		sprintf(fname, "output/best_pop_run%d.out", runnr);
		fpt3 = fopen(fname, "w");
		sprintf(fname, "output/all_pop_run%d.out", runnr);
		fpt4 = fopen(fname, "w");
		sprintf(fname, "output/params_run%d.out", runnr);
		fpt5 = fopen(fname, "w");
		fprintf(fpt1,"# This file contains the data of initial population\n");
		fprintf(fpt2,"# This file contains the data of final population\n");
		fprintf(fpt3,"# This file contains the data of final feasible population (if found)\n");
		fprintf(fpt4,"# This file contains the data of all generations\n");
		fprintf(fpt5,"# This file contains information about inputs as read by the program\n");
		/* -------------------------------- Write files ---------------------------------------- */
		fprintf(fpt5,"\n Population size = %d",popsize);
		fprintf(fpt5,"\n Number of generations = %d",ngen);
		fprintf(fpt5,"\n Number of objective functions = %d",nobj);
		fprintf(fpt5,"\n Number of constraints = %d",ncon);
		fprintf(fpt5,"\n Number of real variables = %d",nreal);
		if (nreal!=0)
		{
			for (i=0; i<nreal; i++)
			{
				fprintf(fpt5,"\n Lower limit of real variable %d = %e",i+1,min_realvar[i]);
				fprintf(fpt5,"\n Upper limit of real variable %d = %e",i+1,max_realvar[i]);
			}
			fprintf(fpt5,"\n Probability of crossover of real variable = %e",pcross_real);
			fprintf(fpt5,"\n Probability of mutation of real variable = %e",pmut_real);
			fprintf(fpt5,"\n Distribution index for crossover = %e",eta_c);
			fprintf(fpt5,"\n Distribution index for mutation = %e",eta_m);
		}
		fprintf(fpt5,"\n Number of binary variables = %d",nbin);
		if (nbin!=0)
		{
			for (i=0; i<nbin; i++)
			{
				fprintf(fpt5,"\n Number of bits for binary variable %d = %d",i+1,nbits[i]);
				fprintf(fpt5,"\n Lower limit of binary variable %d = %e",i+1,min_binvar[i]);
				fprintf(fpt5,"\n Upper limit of binary variable %d = %e",i+1,max_binvar[i]);
			}
			fprintf(fpt5,"\n Probability of crossover of binary variable = %e",pcross_bin);
			fprintf(fpt5,"\n Probability of mutation of binary variable = %e",pmut_bin);
		}
		fprintf(fpt5,"\n Seed for random number generator = %e",seed);
		fprintf(fpt5,"\n Efficiency definition = %d",efficiency);
		fprintf(fpt5,"\n Restriction = %d",restr);
		fprintf(fpt5,"\n app_restr_all_fronts = %d",app_restr_all_fronts);
		fprintf(fpt5,"\n Objective function i = %d",troff_i);
		fprintf(fpt5,"\n Objective function j = %d",troff_j);
		fprintf(fpt5,"\n delta = %e",delta);
		fprintf(fpt5,"\n Geoffrion/tradeoff bound M = %e",m_bound);
		fprintf(fpt5,"\n gamma = %e",gamma);
		fprintf(fpt5,"\n ideal vector = ");
		for (i=0; i<nobj; i++)
		{
			fprintf(fpt5,"%e ",ideal[i]);
		}
		fprintf(fpt5,"\n Number of runs = %d",runs);
		bitlength = 0;
		if (nbin!=0)
		{
			for (i=0; i<nbin; i++)
			{
				bitlength += nbits[i];
			}
		}
		fprintf(fpt1,"# of objectives = %d, # of constraints = %d, # of real_var = %d, # of bits of bin_var = %d, constr_violation, rank, crowding_distance\n",nobj,ncon,nreal,bitlength);
		fprintf(fpt2,"# of objectives = %d, # of constraints = %d, # of real_var = %d, # of bits of bin_var = %d, constr_violation, rank, crowding_distance\n",nobj,ncon,nreal,bitlength);
		fprintf(fpt3,"# of objectives = %d, # of constraints = %d, # of real_var = %d, # of bits of bin_var = %d, constr_violation, rank, crowding_distance\n",nobj,ncon,nreal,bitlength);
		fprintf(fpt4,"# of objectives = %d, # of constraints = %d, # of real_var = %d, # of bits of bin_var = %d, constr_violation, rank, crowding_distance\n",nobj,ncon,nreal,bitlength);


		/* ------------------------------------------ Start algorithm --------------------------------------- */
		nbinmut = 0;
		nrealmut = 0;
		nbincross = 0;
		nrealcross = 0;
		parent_pop = (population *)malloc(sizeof(population));
		child_pop = (population *)malloc(sizeof(population));
		mixed_pop = (population *)malloc(sizeof(population));
		f0_pop = (population *)malloc(sizeof(population));
		allocate_memory_pop (parent_pop, popsize);
		allocate_memory_pop (child_pop, popsize);
		allocate_memory_pop (mixed_pop, 2*popsize);
		allocate_memory_pop (f0_pop, 2*popsize); /* Front0 population */
		/* randomize(); */
		initialize_pop (parent_pop);
		printf("\n Initialization done, now performing first generation");
		decode_pop(parent_pop);
		evaluate_pop (parent_pop);
		assign_rank_and_crowding_distance (parent_pop);
		report_pop (parent_pop, fpt1);
		fprintf(fpt4,"# gen = 1\n");
		report_pop(parent_pop,fpt4);
		printf("\n gen = 1");
		fflush(stdout);
		if (choice!=0)    onthefly_display (parent_pop,gp,1);
		fflush(fpt1);
		fflush(fpt2);
		fflush(fpt3);
		fflush(fpt4);
		fflush(fpt5);
		sleep(1);
		for (i=2; i<=ngen; i++)
		{
			selection (parent_pop, child_pop);
			mutation_pop (child_pop);
			decode_pop(child_pop);
			evaluate_pop(child_pop);
			merge (parent_pop, child_pop, mixed_pop);
			fill_nondominated_sort (mixed_pop, parent_pop, f0_pop);
			/* Comment following four lines if information for all
			generations is not desired, it will speed up the execution */
			/* fprintf(fpt4,"# gen = %d\n",i);
			report_pop(parent_pop,fpt4);

			fflush(fpt4); */
			if (choice!=0)    onthefly_display (parent_pop,gp,i);
			if ((i%50)==0)
				printf("\n gen = %d",i);
		}
		printf("\n Generations of run %d finished, now reporting solutions", runnr);
		report_pop(parent_pop,fpt2);
		report_feasible(parent_pop,fpt3);
		if (nreal!=0)
		{
			fprintf(fpt5,"\n Number of crossover of real variable = %d",nrealcross);
			fprintf(fpt5,"\n Number of mutation of real variable = %d",nrealmut);
		}
		if (nbin!=0)
		{
			fprintf(fpt5,"\n Number of crossover of binary variable = %d",nbincross);
			fprintf(fpt5,"\n Number of mutation of binary variable = %d",nbinmut);
		}
		fflush(stdout);
		fflush(fpt1);
		fflush(fpt2);
		fflush(fpt3);
		fflush(fpt4);
		fflush(fpt5);
		fclose(fpt1);
		fclose(fpt2);
		fclose(fpt3);
		fclose(fpt4);
		fclose(fpt5);
		if (choice!=0)
		{
			pclose(gp);
		}

		deallocate_memory_pop (parent_pop, popsize);
		deallocate_memory_pop (child_pop, popsize);
		deallocate_memory_pop (mixed_pop, 2*popsize);
		deallocate_memory_pop (f0_pop, 2*popsize);
		free (parent_pop);
		free (child_pop);
		free (mixed_pop);
		free (f0_pop);
    }

	if (nreal!=0)
	{
		free (min_realvar);
		free (max_realvar);
	}
	if (nbin!=0)
	{
		free (min_binvar);
		free (max_binvar);
		free (nbits);
	}
    printf("\n Routine successfully exited \n");
    return (0);
}
